library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity EvoBlock is
	generic (NUM_ARRAYS	: integer := 8;
				NUM_COLS		: integer := 4);
	port (	clk			: in	std_logic;
				reset			: in	std_logic;

				arrayNum		: in	integer range 0 to (NUM_ARRAYS-1);
				cellNum		: in	integer range 0 to ((NUM_COLS*8)-1);
				configDIn	: in	std_logic_vector(19 downto 0);
				configWEn	: in	std_logic;

				ramDIn		: in	std_logic_vector(7 downto 0);
				ramWEn		: in	std_logic;
				ramDOut		: out	std_logic_vector(7 downto 0);

				vector		: in	std_logic_vector(7 downto 0);
				mask			: in	std_logic_vector(7 downto 0);

				output		: out std_logic_vector(((NUM_ARRAYS*8)-1) downto 0);
				match			: out std_logic_vector(((NUM_ARRAYS*8)-1) downto 0));
end EvoBlock;

architecture Arch of EvoBlock is

	component CellArray is
		generic ( NUM_COLS	: integer := 8);
		port (	clk			: in	std_logic;
					reset			: in	std_logic;
					dIn			: in	std_logic_vector(7 downto 0);
					dOut			: out	std_logic_vector(7 downto 0);
					cellNum		: in	integer range 0 to ((NUM_COLS * 8)-1);
					configDIn	: in	std_logic_vector(19 downto 0);
					configWEn	: in	std_logic);
	end component;

	type RAM_TYPE is array (255 downto 0) of std_logic_vector(7 downto 0);

	signal targetRAM			: RAM_TYPE;
	signal targetRAMOut		: std_logic_vector(7 downto 0);

begin

ramDOut <= targetRAMOut;

-- TargetRAM --------------------------------------------------------------------------------------------------
-- Target truth table
---------------------------------------------------------------------------------------------------------------
process (clk)
begin
	if (clk'event and clk='1') then
		if (ramWEn = '1') then
			targetRAM(to_integer(unsigned(vector))) <= ramDIn;
			targetRAMOut <= ramDIn;
		else
			targetRAMOut <= targetRAM(to_integer(unsigned(vector)));
		end if;
	end if;
end process;

-- ArrayGenerate ----------------------------------------------------------------------------------------------
-- Create a block of 8 Cell Arrays
---------------------------------------------------------------------------------------------------------------
ArrayGenerate : for arrayNumber in 0 to (NUM_ARRAYS-1) generate

	signal arrayConfigWEn	: std_logic;
	signal arrayOut			: std_logic_vector(7 downto 0);
	signal arrayOutReg		: std_logic_vector(7 downto 0);

begin

	arrayConfigWEn <= '1' when (arrayNumber=arrayNum and configWEn='1') else '0';

	CellArrayInst : CellArray
	generic map(NUM_COLS			=> NUM_COLS)
	port map (	clk				=> clk,
					reset				=> reset,
					dIn				=> vector,
					dOut				=> arrayOut,
					cellNum			=>	cellNum,
					configDIn		=> configDIn,
					configWEn		=> arrayConfigWEn);

	process (clk)
	begin
		if (clk'event and clk='1') then
			arrayOutReg <= arrayOut;
		end if;
	end process;

	output((arrayNumber*8)+7 downto (arrayNumber*8)) <= arrayOutReg;
	match((arrayNumber*8)+7 downto (arrayNumber*8)) <= (arrayOutReg xnor targetRAMOut) and mask;

end generate;

end Arch;
