------------------------------------------------------------------------------
--				Width		Write				Read
-- Reg 0		(3)		arrayNum			arrayNum						+ 0
-- Reg 1		(6)		cellNum			cellNum						+ 4
-- Reg 2		(20)		config											+ 8
-- Reg 3		(8)		vector			vector						+ 12
-- Reg 4		(8)		mask				mask							+ 16
-- Reg 5		(8)		ramData			ramData						+ 20
-- Reg 6		(32)							output (3 downto 0)		+ 24
-- Reg 7		(32)							output (7 downto 4)		+ 28
-- Reg 8		(32)							match (3 downto 0)		+ 32
-- Reg 9		(32)							match (7 downto 4)		+ 36
-- Reg 10	(4)		DebugLEDS										+ 40
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity opb_evoblock is
generic
(
	C_BASEADDR		: std_logic_vector	:= X"00000000";
	C_HIGHADDR		: std_logic_vector	:= X"0000000F";
	C_OPB_AWIDTH	: integer				:= 32;
	C_OPB_DWIDTH	: integer				:= 32;
	C_USER_ID_CODE	: integer				:= 3;
	C_FAMILY			: string					:= "virtex2p"
);
port
(
	ledDebug			: out	std_logic_vector(3 downto 0);

	OPB_Clk			: in	std_logic;
	OPB_Rst			: in	std_logic;
	Sl_DBus			: out	std_logic_vector(0 to C_OPB_DWIDTH-1);
	Sl_errAck		: out	std_logic;
	Sl_retry			: out	std_logic;
	Sl_toutSup		: out	std_logic;
	Sl_xferAck		: out	std_logic;
	OPB_ABus			: in	std_logic_vector(0 to C_OPB_AWIDTH-1);
	OPB_BE			: in	std_logic_vector(0 to C_OPB_DWIDTH/8-1);
	OPB_DBus			: in	std_logic_vector(0 to C_OPB_DWIDTH-1);
	OPB_RNW			: in	std_logic;
	OPB_select		: in	std_logic;
	OPB_seqAddr		: in	std_logic
);

	attribute SIGIS						: string;
	attribute SIGIS of OPB_Clk			: signal is "Clk";
	attribute SIGIS of OPB_Rst			: signal is "Rst";

end entity opb_evoblock;

------------------------------------------------------------------------------
-- Architecture section
------------------------------------------------------------------------------

architecture IMP of opb_evoblock is

	component EvoBlock is
		generic (NUM_ARRAYS	: integer := 8;
					NUM_COLS		: integer := 5);
		port (	clk			: in	std_logic;
					reset			: in	std_logic;
					arrayNum		: in	integer range 0 to (NUM_ARRAYS-1);
					cellNum		: in	integer range 0 to ((NUM_COLS*8)-1);
					configDIn	: in	std_logic_vector(19 downto 0);
					configWEn	: in	std_logic;
					ramDIn		: in	std_logic_vector(7 downto 0);
					ramWEn		: in	std_logic;
					ramDOut		: out	std_logic_vector(7 downto 0);
					vector		: in	std_logic_vector(7 downto 0);
					mask			: in	std_logic_vector(7 downto 0);
					output		: out std_logic_vector(((NUM_ARRAYS*8)-1) downto 0);
					match			: out std_logic_vector(((NUM_ARRAYS*8)-1) downto 0));
	end component;

	signal data_BUS2IP		: std_logic_vector(31 downto 0);
	signal data_IP2BUS		: std_logic_vector(31 downto 0);

	signal chipSel				: std_logic;

	signal arrayNumReg		: std_logic_vector(2 downto 0);
	signal cellNumReg			: unsigned(5 downto 0);
	signal configWEn			: std_logic;
	signal ramWEn				: std_logic;
	signal ramDOut				: std_logic_vector(7 downto 0);
	signal vectorReg			: unsigned(7 downto 0);
	signal maskReg				: std_logic_vector(7 downto 0);
	signal output				: std_logic_vector(63 downto 0);
	signal match				: std_logic_vector(63 downto 0);

	signal arrayNum			: integer range 0 to 8;
	signal cellNum				: integer range 0 to 39;
	signal vector				: std_logic_vector(7 downto 0);

	signal ledReg				: std_logic_vector(3 downto 0);

	function reverseVector (a: in std_logic_vector) return std_logic_vector is
		variable result: std_logic_vector(a'reverse_range);
	begin
		for i in a'range loop
			result((a'length-1)-i) := a(i);
 		end loop;
		return result;
	end;

begin

ledDebug		<= not ledReg;

chipSel		<= '1' when (OPB_select='1' and OPB_ABus(0 to 25)=C_BASEADDR(0 to 25)) else '0';

Sl_DBus		<= reverseVector(data_IP2BUS);
data_BUS2IP	<= reverseVector(OPB_DBus);

Sl_errAck	<= '0';
Sl_retry		<= '0';
Sl_toutSup	<= '0';

-- Convert types for connection to EvoBlock
arrayNum	<= to_integer(unsigned(arrayNumReg));
cellNum	<= to_integer(cellNumReg);
vector	<= std_logic_vector(vectorReg);

EvoBlk : EvoBlock
generic map(NUM_ARRAYS	=> 8,
				NUM_COLS		=> 5)
port map(	clk			=> OPB_Clk,
				reset			=> OPB_Rst,
				arrayNum		=> arrayNum,
				cellNum		=> cellNum,
				configDIn	=> data_BUS2IP(19 downto 0),
				configWEn	=> configWEn,
				ramDIn		=> data_BUS2IP(7 downto 0),
				ramWEn		=> ramWEn,
				ramDOut		=> ramDOut,
				vector		=> vector,
				mask			=> maskReg,
				output		=> output,
				match			=> match);

ReadWriteControl : process(OPB_RNW, chipSel, OPB_ABus, arrayNumReg, cellNumReg, vectorReg, maskReg, ramDOut, output, match)
	variable data_IP2BUS_var	: std_logic_vector(31 downto 0);
	variable Sl_xferAck_var		: std_logic;
	variable ramWEn_var			: std_logic;
	variable configWEn_var		: std_logic;
begin

	data_IP2BUS_var	:= (others => '0');
	Sl_xferAck_var		:= '0';
	configWEn_var		:= '0';
	ramWEn_var			:= '0';

	-- Read
	if (OPB_RNW='1' and chipSel='1') then
		case OPB_ABus(26 to 29) is
			when x"0"	=> data_IP2BUS_var(2 downto 0)	:= arrayNumReg;
			when x"1"	=> data_IP2BUS_var(5 downto 0)	:= std_logic_vector(cellNumReg);
			when x"3"	=> data_IP2BUS_var(7 downto 0)	:= std_logic_vector(vectorReg);
			when x"4"	=> data_IP2BUS_var(7 downto 0)	:= maskReg;
			when x"5"	=> data_IP2BUS_var(7 downto 0)	:= ramDOut;
			when x"6"	=> data_IP2BUS_var					:= output(31 downto 0);
			when x"7"	=> data_IP2BUS_var					:= output(63 downto 32);
			when x"8"	=> data_IP2BUS_var					:= match(31 downto 0);
			when x"9"	=> data_IP2BUS_var					:= match(63 downto 32);
			when others	=> null;
		end case;
		Sl_xferAck_var	:= '1';
	end if;

	-- Write
	if (OPB_RNW='0' and chipSel='1') then
		case OPB_ABus(26 to 29) is
			when x"2"	=> configWEn_var	:= '1';
			when x"5"	=> ramWEn_var		:= '1';
			when others	=> null;
		end case;
		Sl_xferAck_var	:= '1';
	end if;

	data_IP2BUS		<= data_IP2BUS_var;
	Sl_xferAck		<= Sl_xferAck_var;
	configWEn		<= configWEn_var;
	ramWEn			<= ramWEn_var;

end process ReadWriteControl;

WriteControl : process(OPB_Clk)
begin
	if (OPB_Clk'event and OPB_Clk='1') then
		if (OPB_Rst='1') then
			arrayNumReg <= (others => '0');
			cellNumReg <= (others => '0');
			vectorReg <= (others => '0');
			maskReg <= (others => '0');
			ledReg <= (others => '0');

		elsif (OPB_RNW='0' and chipSel='1') then
			case OPB_ABus(26 to 29) is
				when x"0"	=> arrayNumReg	<= data_BUS2IP(2 downto 0);
				when x"1"	=> cellNumReg	<= unsigned(data_BUS2IP(5 downto 0));
				when x"2"	=> cellNumReg	<= cellNumReg + 1;
				when x"3"	=> vectorReg	<= unsigned(data_BUS2IP(7 downto 0));
				when x"4"	=> maskReg		<= data_BUS2IP(7 downto 0);
				when x"5"	=> vectorReg	<= vectorReg + 1;
				when x"A"	=> ledReg		<= data_BUS2IP(3 downto 0);
				when others	=> null;
			end case;

		end if;
	end if;
end process WriteControl;

end IMP;
